@@ -3,7 +3,7 @@
   missing.h - prototype for *.c in ./missing, and
   	      for missing timeval struct
 
-  $Author: nobu $
+  $Author: usa $
   created at: Sat May 11 23:46:03 JST 2002
 
 ************************************************/
@@ -162,6 +162,8 @@ RUBY_EXTERN const union bytesequence4_or_float rb_nan;
 #    include <ieeefp.h>
 #    endif
 #  define isinf(x) (!finite(x) && !isnan(x))
+#  elif defined(__cplusplus) && __cplusplus >= 201103L
+#    include <cmath> // it must include constexpr bool isinf(double);
 #  else
 RUBY_EXTERN int isinf(double);
 #  endif
@@ -170,7 +172,11 @@ RUBY_EXTERN int isinf(double);
 
 #ifndef isnan
 # ifndef HAVE_ISNAN
+#  if defined(__cplusplus) && __cplusplus >= 201103L
+#    include <cmath> // it must include constexpr bool isnan(double);
+#  else
 RUBY_EXTERN int isnan(double);
+#  endif
 # endif
 #endif
 
